"Implementation of tzinfo classes for use with datetime.datetime."

from __future__ import unicode_literals

import time
from datetime import timedelta, tzinfo

from django.utils.encoding import (
    DEFAULT_LOCALE_ENCODING, force_str, force_text,
)

from django_compat_patcher.deprecation import *

warnings.warn(
    "django.utils.tzinfo will be removed in Django 1.9. "
    "Use django.utils.timezone instead.",
    RemovedInDjango19Warning, stacklevel=2)


# Python's doc say: "A tzinfo subclass must have an __init__() method that can
# be called with no arguments". FixedOffset and LocalTimezone don't honor this
# requirement. Defining __getinitargs__ is sufficient to fix copy/deepcopy as
# well as pickling/unpickling.

class FixedOffset(tzinfo):
    "Fixed offset in minutes east from UTC."

    def __init__(self, offset):
        warnings.warn(
            "django.utils.tzinfo.FixedOffset will be removed in Django 1.9. "
            "Use django.utils.timezone.get_fixed_timezone instead.",
            RemovedInDjango19Warning)
        if isinstance(offset, timedelta):
            self.__offset = offset
            offset = self.__offset.seconds // 60
        else:
            self.__offset = timedelta(minutes=offset)

        sign = '-' if offset < 0 else '+'
        self.__name = "%s%02d%02d" % (sign, abs(offset) / 60., abs(offset) % 60)

    def __repr__(self):
        return self.__name

    def __getinitargs__(self):
        return self.__offset,

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return timedelta(0)


# This implementation is used for display purposes. It uses an approximation
# for DST computations on dates >= 2038.

# A similar implementation exists in django.utils.timezone. It's used for
# timezone support (when USE_TZ = True) and focuses on correctness.

class LocalTimezone(tzinfo):
    "Proxy timezone information from time module."

    def __init__(self, dt):
        warnings.warn(
            "django.utils.tzinfo.LocalTimezone will be removed in Django 1.9. "
            "Use django.utils.timezone.get_default_timezone instead.",
            RemovedInDjango19Warning)
        tzinfo.__init__(self)
        self.__dt = dt
        self._tzname = self.tzname(dt)

    def __repr__(self):
        return force_str(self._tzname)

    def __getinitargs__(self):
        return self.__dt,

    def utcoffset(self, dt):
        if self._isdst(dt):
            return timedelta(seconds=-time.altzone)
        else:
            return timedelta(seconds=-time.timezone)

    def dst(self, dt):
        if self._isdst(dt):
            return timedelta(seconds=-time.altzone) - timedelta(seconds=-time.timezone)
        else:
            return timedelta(0)

    def tzname(self, dt):
        is_dst = False if dt is None else self._isdst(dt)
        try:
            return force_text(time.tzname[is_dst], DEFAULT_LOCALE_ENCODING)
        except UnicodeDecodeError:
            return None

    def _isdst(self, dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        try:
            stamp = time.mktime(tt)
        except (OverflowError, ValueError):
            # 32 bit systems can't handle dates after Jan 2038, and certain
            # systems can't handle dates before ~1901-12-01:
            #
            # >>> time.mktime((1900, 1, 13, 0, 0, 0, 0, 0, 0))
            # OverflowError: mktime argument out of range
            # >>> time.mktime((1850, 1, 13, 0, 0, 0, 0, 0, 0))
            # ValueError: year out of range
            #
            # In this case, we fake the date, because we only care about the
            # DST flag.
            tt = (2037,) + tt[1:]
            stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0



from __future__ import absolute_import, print_function, unicode_literals

from functools import partial

from ..registry import register_compatibility_fixer

# for backward-compatibility fixers
django1_11_bc_fixer = partial(register_compatibility_fixer,
                              fixer_reference_version="1.11",
                              fixer_applied_from_django="1.11")



@django1_11_bc_fixer()
def fix_behaviour_widget_build_attrs(utils):
    """
    Preserve the "future" templatetags library, with its improved `firstof` and `cycle` tags.
    """
    from django.forms.widgets import Widget

    # original_build_attrs = Widget.build_attrs  UNUSED

    def build_attrs(self, base_attrs=None, extra_attrs=None, **kwargs):
        if kwargs:
            # old version
            assert not extra_attrs
            attrs = dict(self.attrs, **kwargs)
            if base_attrs:  # actually the old "extra_attrs"
                attrs.update(base_attrs)
        else:
            # new version
            attrs = base_attrs.copy()
            if extra_attrs is not None:
                attrs.update(extra_attrs)
        return attrs
    utils.inject_callable(Widget, "build_attrs", build_attrs)

'''
def build_attrs(self, extra_attrs=None, **kwargs):
    "Helper function for building an attribute dictionary."
    attrs = dict(self.attrs, **kwargs)
    if extra_attrs:
        attrs.update(extra_attrs)
    return attrs

def build_attrs(self, base_attrs, extra_attrs=None):
    "Helper function for building an attribute dictionary."
    attrs = base_attrs.copy()
    if extra_attrs is not None:
        attrs.update(extra_attrs)
    return attrs
'''



from __future__ import absolute_import, print_function, unicode_literals

from functools import partial

from ..deprecation import *
from ..registry import register_compatibility_fixer


# for backward-compatibility fixers
django1_8_bc_fixer = partial(register_compatibility_fixer,
                             fixer_reference_version="1.8",
                             fixer_applied_from_django="1.8")



@django1_8_bc_fixer()
def fix_outsourcing_contrib_comments(utils):
    """
    Keep 'django.contrib.comments' as an import alias for the now external package
    'django_comments' (django-contrib-comments on pypi) ; the latter must be installed separately.
    """
    utils.register_import_alias(alias_name="django.contrib.comments",
                                real_name="django_comments")



from __future__ import absolute_import, print_function, unicode_literals

from functools import partial

from ..deprecation import *
from ..registry import register_compatibility_fixer


# for backward-compatibility fixers
django1_7_bc_fixer = partial(register_compatibility_fixer,
                             fixer_reference_version="1.7",
                             fixer_applied_from_django="1.7")





from __future__ import absolute_import, print_function, unicode_literals

import functools
from . import fixers, utilities, deprecation, registry

__APPLIED_FIXERS = set()


def patch(settings=None):
    """Patches the Django package with relevant fixers.
    
    A settings dict/objects can be provided, to REPLACE lookups in Django settings.
    
    Returns a list of ids of fixers applied.
    """

    utilities.apply_runtime_settings(settings)  # called even if settings are empty

    # print("Fixers are:", registry.FIXERS_REGISTRY)
    django_version = utilities.get_django_version()
    log = functools.partial(utilities.emit_log, level="DEBUG")
    relevant_fixers = registry.get_relevant_fixers(current_django_version=django_version,
                                                   settings=settings,
                                                   log=log)

    # REVERSED is better for backwards compatibility, for now...
    relevant_fixers.sort(key=lambda x:x["fixer_reference_version"], reverse=True)

    pre_fixers = [f for f in relevant_fixers if not f["fixer_delayed"]]
    post_fixers = [f for f in relevant_fixers if f["fixer_delayed"]]
    assert len(relevant_fixers) == len(pre_fixers) + len(post_fixers)

    def _apply_fixers(fixers):
        for fixer in fixers:
            # print("Applying fixer", fixer)
            # TODO - create custom injected "utils" object with context information, logging, warnings, etc.
            if fixer['fixer_id'] not in __APPLIED_FIXERS:
                utilities.emit_log("Django compat fixer '{}-{}' is getting applied".format(
                    fixer["fixer_family"], fixer['fixer_id']), level="INFO"
                )
                try:
                    fixer["fixer_callable"](utilities)
                    __APPLIED_FIXERS.add(fixer['fixer_id'])
                except utilities.SkipFixerException as e:
                    utilities.emit_log("Django compat fixer '{}-{}' was not applied, reason: {}".format(
                        fixer["fixer_family"], fixer['fixer_id'], e), level="WARNING"
                    )
            else:
                utilities.emit_log("Django compat fixer '{}' was already applied".
                                         format(fixer['fixer_id']), level="WARNING")

    _apply_fixers(pre_fixers)
    import django
    django.setup()  # theoretically idempotent (except regarding logging?)
    _apply_fixers(post_fixers)

    return relevant_fixers



"""
This module allows to create aliases between packages, so that one
can import a module under a different name (mainly for retrocompatibility purpose).

Note that it is NOT about aliasing local references to modules after import, eg. with "from package import module_name as other_module_name".
Here we really deal with aliasing the "full name" of the module, as it will appear in sys.modules.

There are different ways of implementing "import aliases":

- overridding __import__ (difficult, since this API loads parent modules too)
- using a custom module loader which creates a wrapper class (with setattr,
  getattr etc...), and injects it as a proxy in sys.modules[alias]
- using a custom module loader, which simply copies a reference to the real module
  into sys.modules (and on recent python version, updates its __spec__ to gives hints about this operation).
  *THIS* is the way it is currently implemented, so that imported modules keep being SINGLETONS whatever their
  possible aliases.

Beware about not creating loops with your aliases, as this could trigger infinite recursions.
"""

import os, sys
import importlib, contextlib


# maps ALIASES to REAL MODULES
MODULES_ALIASES_REGISTRY = []


@contextlib.contextmanager
def enrich_import_error(alias_name):
    try:
        yield
    except ImportError as e:
        #print(vars(e), str(e), dir(e), e.args)
        # ImportError exception has different structure between py2k and py3k
        enrich_message = lambda msg: "%s (when loading alias name '%s')" % (msg, alias_name)
        if hasattr(e, "msg"):  # py3k
            e.msg = enrich_message(e.msg)
        else:  # py2k, the "message" attribute is actually ignored by __str__
            e.args = (enrich_message(e.args[0]),)
        raise


def register_module_alias(alias_name, real_name):
    assert not alias_name.startswith("."), alias_name
    assert not real_name.startswith("."), real_name
    assert alias_name != real_name, alias_name  # lots of other import cycles are possible though
    entry = (alias_name, real_name)
    if entry not in MODULES_ALIASES_REGISTRY:
        MODULES_ALIASES_REGISTRY.append(entry)
        return True
    return False


def _get_module_alias_real_name(fullname):
    """
    returns the real name of module (when fullname is an alias name)
    or None.
    """
    for k, v in MODULES_ALIASES_REGISTRY:
        if (k == fullname) or fullname.startswith(k +"."):
            return fullname.replace(k, v)
    return None


try:

    import importlib.machinery, importlib.abc

except ImportError:

    # OLD STYLE : we use PEP 302 find_module/load_module API, available for python2.7+

    is_new_style_proxifier = False


    class AliasingLoader(object):

        def __init__(self, real_name, alias_name):
            self.real_name = real_name
            self.alias_name = alias_name

        def load_module(self, name):
            if name in sys.modules:
                return sys.modules[name]  # shortcut
            # we let the standard machinery handle sys.modules, module attrs etc.
            with enrich_import_error(self.alias_name):
                module = importlib.import_module(self.real_name, package=None)
            sys.modules[name] = module  # cached
            return module


    class ModuleAliasFinder(object):

        @classmethod
        def find_module(self, fullname, *args, **kwargs):

            #print("OldMetaPathFinder FINDMODULE", fullname, args, kwargs)

            real_name = _get_module_alias_real_name(fullname)
            if real_name is None:
                return None  # no aliased module is known
            return AliasingLoader(real_name=real_name, alias_name=fullname)


else:

    # NEW STYLE : we use modern import hooks (PEP 451 etc.)

    is_new_style_proxifier = True


    class AliasingLoader(importlib.abc.Loader):

        target_spec_backup = None

        def __init__(self, real_name, alias_name):
            self.real_name = real_name
            self.alias_name = alias_name

        def create_module(self, spec):
            # we do the real loading of aliased module here
            with enrich_import_error(self.alias_name):
                module = importlib.import_module(self.real_name, package=None)
            assert module.__name__ == self.real_name, module.__name__
            self.target_spec_backup = module.__spec__
            return module

        def exec_module(self, module):
            # __name__ is, on some python versions, overridden by init_module_attrs(_force_name=True)
            assert module.__name__ in (self.alias_name, self.real_name), module.__name__
            module.__name__ = self.real_name
            assert module.__spec__.origin == "alias", module.__spec__  # well overridden
            assert module.__spec__.loader_state["aliased_spec"] is None
            assert self.target_spec_backup, self.target_spec_backup
            module.__spec__.loader_state["aliased_spec"] = self.target_spec_backup
            pass  # nothing else to do, module already loaded


    class ModuleAliasFinder(importlib.abc.MetaPathFinder):
        """
        Note : due to new call of _init_module_attrs(spec...) on aliased
        module, its __spec__ attribute will be changed by this new import...
        """

        @classmethod
        def find_spec(cls, fullname, *args, **kwargs):

            #print("MetaPathFinder FINDSPEC", fullname, args, kwargs)

            real_name = _get_module_alias_real_name(fullname)
            if real_name is None:
                return None  # no aliased module is known

            # TODO emit warnings/logging

            alias_loader = AliasingLoader(real_name=real_name, alias_name=fullname)

            spec = importlib.machinery.ModuleSpec(name=fullname,
                                           loader=alias_loader,
                                           origin="alias",
                                           loader_state={"aliased_name": fullname,
                                                         "aliased_spec": None},
                                           is_package=True)  # in doubt, assume...

            return spec



def install_import_proxifier():
    """
    Add a meta path hook before all others, so that new module loadings
    may be redirected to aliased module.

    Idempotent function.
    """
    if ModuleAliasFinder not in sys.meta_path:
        sys.meta_path.insert(0, ModuleAliasFinder)
    assert ModuleAliasFinder in sys.meta_path, sys.meta_path



if __name__ == "__main__":

    import logging.handlers

    install_import_proxifier()
    install_import_proxifier()  # idempotent

    register_module_alias("json.tool_alias", "json.tool")
    register_module_alias("mylogging", "logging")
    register_module_alias("mylogging", "logging")
    register_module_alias("infinite_recursion", "infinite_recursion2")
    register_module_alias("infinite_recursion2", "infinite_recursion")
    register_module_alias("unexisting_alias", "unexisting_module")

    fullname = "logging.handlers"
    print ("ALREADY THERE:", fullname, fullname in sys.modules)

    import json.tool_alias
    assert sys.modules["json.tool_alias"]
    assert json.tool_alias

    fullname = "json.comments"
    print ("ALREADY THERE:", fullname, fullname in sys.modules)

    fullname = "json.tool"
    print ("ALREADY THERE:", fullname, fullname in sys.modules)

    import json.tool
    if is_new_style_proxifier:
        print ("JSON TOOL SPEC loader_state:", json.tool.__spec__.loader_state)

    assert json.tool_alias is json.tool, (json.tool_alias,  json.tool)
    assert json.tool.__name__ == "json.tool", json.tool.__name__
    if is_new_style_proxifier:
        assert json.tool.__spec__.origin == "alias", json.tool.__spec__.origin
        assert json.tool.__spec__.name == "json.tool_alias"
        assert json.tool.__spec__.loader_state["aliased_spec"].name == "json.tool"

    from mylogging import config
    assert config.dictConfig
    assert "logging.config" in sys.modules

    from mylogging.handlers import RotatingFileHandler
    from logging.handlers import RotatingFileHandler as RotatingFileHandlerOriginal
    print("mylogging led to RotatingFileHandler", RotatingFileHandler)
    assert RotatingFileHandler is RotatingFileHandlerOriginal

    try:
        import infinite_recursion
    except RuntimeError:
        pass
    else:
        raise RuntimeError("Import should have led to infinite recursion")

    try:
        import unexisting_alias
    except ImportError as e:
        assert "unexisting_alias" in str(e), (str(e), vars(e))
        assert "unexisting_module" in str(e), str(e)
    else:
        raise RuntimeError("import error noy raised")

    try:
        import unexisting_module
    except ImportError as e:
        assert "unexisting_alias" not in str(e), str(e)
        assert "unexisting_module" in str(e), str(e)
    else:
        raise RuntimeError("import error noy raised")

    print("\n*** Tests were successful ***")



from __future__ import absolute_import, print_function, unicode_literals

import collections

from django.utils import six

from django_compat_patcher.utilities import _tuplify_version
from . import utilities

FIXERS_REGISTRY = collections.OrderedDict()


def _extract_doc(func):
    """Extract and check the docstring of a callable"""
    doc = func.__doc__
    assert doc, "Fixer %r must provide a help string" % func
    return doc


def register_compatibility_fixer(fixer_reference_version,
                                 fixer_applied_from_django=None,  # INCLUDING this version
                                 fixer_applied_upto_django=None,  # EXCLUDING this version
                                 fixer_delayed=False,
                                 feature_supported_from_django=None,
                                 feature_supported_upto_django=None,
                                 ):
    """
    Registers a fixer, which will be executed only if current django version
    is >= `fixer_applied_from_django` and < `fixer_applied_upto_django` (let them None to have no limit).

    `fixer_reference_version` is mainly here for documentation purposes.

    `feature_supported_from_django` and `feature_supported_upto_django` will be used to limit
    range of django versions for which related unit-tests are expected to work (i.e versions
    for which the feature is available, either as a monkey-paching or as standard django code).

    Version identifiers must be strings, eg. "1.9.1".

    If `fixer_delayed`is True, the fixer is only applied after django.setup() has been called
    (eg. so that ORM models are available for introspection and patching).
    """

    assert isinstance(fixer_reference_version, six.string_types)  # eg. "1.9"
    assert fixer_delayed in (True, False), fixer_delayed

    fixer_family = "django" + fixer_reference_version  # eg. django1.9

    def _register_simple_fixer(func):
        fixer_id = func.__name__  # untouched ATM, not fully qualified
        new_fixer = dict(fixer_callable=func,
                         fixer_explanation=_extract_doc(func),
                         fixer_id=fixer_id,
                         fixer_reference_version=_tuplify_version(fixer_reference_version),
                         fixer_family=fixer_family,
                         fixer_applied_from_django=_tuplify_version(fixer_applied_from_django),
                         fixer_applied_upto_django=_tuplify_version(fixer_applied_upto_django),
                         fixer_delayed=fixer_delayed,
                         feature_supported_from_django=_tuplify_version(feature_supported_from_django),
                         feature_supported_upto_django=_tuplify_version(feature_supported_upto_django), )

        assert fixer_id not in FIXERS_REGISTRY, "duplicate fixer id %s detected" % fixer_id
        FIXERS_REGISTRY[fixer_id] = new_fixer
        # print("FIXERS_REGISTRY", FIXERS_REGISTRY)

    return _register_simple_fixer


def get_fixer_by_id(fixer_id):
    return FIXERS_REGISTRY[fixer_id]


def get_relevant_fixers(current_django_version,
                        settings=None,
                        log=None):
    """
    Selects the fixers to be applied on the target django version, based on the
    metadata of fixers, but also inclusion/exclusion lists provided as arguments.

    For inclusion/exclusion filters, a special "*" value means "all fixers", 
    else a list of strings is expected.
    """

    current_django_version = _tuplify_version(current_django_version)

    relevant_fixers = []

    log = log or (lambda x: x)

    ALL = "*"  # special value for patcher settings
    include_fixer_ids = utilities.get_patcher_setting("DCP_INCLUDE_FIXER_IDS", settings=settings)
    include_fixer_families = utilities.get_patcher_setting("DCP_INCLUDE_FIXER_FAMILIES", settings=settings)
    exclude_fixer_ids = utilities.get_patcher_setting("DCP_EXCLUDE_FIXER_IDS", settings=settings)
    exclude_fixer_families = utilities.get_patcher_setting("DCP_EXCLUDE_FIXER_FAMILIES", settings=settings)

    for fixer_id, fixer in FIXERS_REGISTRY.items():
        assert fixer_id == fixer["fixer_id"], fixer

        if (fixer["fixer_applied_from_django"] and
                    current_django_version < fixer["fixer_applied_from_django"]):  # SMALLER STRICTLY
            log("Skipping fixer %s, useful only in subsequent django versions" % fixer_id)
            continue

        if (fixer["fixer_applied_upto_django"] and
                    current_django_version >= fixer["fixer_applied_upto_django"]):  # GREATER OR EQUAL
            log("Skipping fixer %s, useful only in previous django versions" % fixer_id)
            continue

        included = False
        if include_fixer_ids == ALL or (fixer_id in include_fixer_ids):
            included = True
        if include_fixer_families == ALL or (fixer["fixer_family"] in include_fixer_families):
            included = True

        if not included:
            log("Skipping fixer having neither id (%s) nor family (%s) included by patcher settings" % (fixer_id, fixer["fixer_family"]))
            continue

        if exclude_fixer_ids == ALL or (fixer_id in exclude_fixer_ids):
            log("Skipping fixer %s, excluded by patcher settings" % fixer_id)
            continue

        if exclude_fixer_families == ALL or (fixer["fixer_family"] in exclude_fixer_families):
            log("Skipping fixer %s, having family %s excluded by patcher settings" % (fixer_id, fixer["fixer_family"]))
            continue

        # cheers, this fixer has passed all filters!
        relevant_fixers.append(fixer)

    return relevant_fixers


def _extract_fixer_ids(fixers):
    return [f["fixer_id"] for f in fixers]


def get_relevant_fixer_ids(*args, **kwargs):
    fixers = get_relevant_fixers(*args, **kwargs)
    return _extract_fixer_ids(fixers)


def get_all_fixers():
    return list(FIXERS_REGISTRY.values())



from __future__ import absolute_import, print_function, unicode_literals

from django_compat_patcher.utilities import emit_warning

class RemovedInDjango21Warning(PendingDeprecationWarning):
    pass


class RemovedInDjango20Warning(DeprecationWarning):
    pass


class RemovedInDjango110Warning(DeprecationWarning):
    pass


class RemovedInDjango19Warning(DeprecationWarning):
    pass


class RemovedInDjango18Warning(DeprecationWarning):
    pass


class RemovedInDjango17Warning(DeprecationWarning):
    pass


class RemovedInDjango16Warning(DeprecationWarning):
    pass


RemovedInNextVersionWarning = RemovedInDjango21Warning


# replacement for "import warnings", which behaves according to DCP settings
class WarningsProxy():
    def warn(self, *args, **kwargs):
        emit_warning(*args, **kwargs)
warnings = WarningsProxy()



import os
import re
import sys
from datetime import datetime

from django.conf import settings

from django.template.base import (
    Node, Template, TemplateSyntaxError,
)

from django_compat_patcher.deprecation import *

def include_is_allowed(filepath, allowed_include_roots):
    filepath = os.path.abspath(filepath)
    for root in allowed_include_roots:
        if filepath.startswith(root):
            return True
    return False


class SsiNode(Node):
    def __init__(self, filepath, parsed):
        self.filepath = filepath
        self.parsed = parsed

    def render(self, context):
        filepath = self.filepath.resolve(context)

        if not include_is_allowed(filepath, context.template.engine.allowed_include_roots):
            if settings.DEBUG:
                return "[Didn't have permission to include file]"
            else:
                return ''  # Fail silently for invalid includes.
        try:
            with open(filepath, 'r') as fp:
                output = fp.read()
        except IOError:
            output = ''
        if self.parsed:
            try:
                t = Template(output, name=filepath, engine=context.template.engine)
                return t.render(context)
            except TemplateSyntaxError as e:
                if settings.DEBUG:
                    return "[Included template had syntax error: %s]" % e
                else:
                    return ''  # Fail silently for invalid included templates.
        return output


#@register.tag
def ssi(parser, token):
    """
    Outputs the contents of a given file into the page.

    Like a simple "include" tag, the ``ssi`` tag includes the contents
    of another file -- which must be specified using an absolute path --
    in the current page::

        {% ssi "/home/html/ljworld.com/includes/right_generic.html" %}

    If the optional "parsed" parameter is given, the contents of the included
    file are evaluated as template code, with the current context::

        {% ssi "/home/html/ljworld.com/includes/right_generic.html" parsed %}
    """
    warnings.warn(
        "The {% ssi %} tag is deprecated. Use the {% include %} tag instead.",
        RemovedInDjango110Warning,
    )

    bits = token.split_contents()
    parsed = False
    if len(bits) not in (2, 3):
        raise TemplateSyntaxError("'ssi' tag takes one argument: the path to"
                                  " the file to be included")
    if len(bits) == 3:
        if bits[2] == 'parsed':
            parsed = True
        else:
            raise TemplateSyntaxError("Second (optional) argument to %s tag"
                                      " must be 'parsed'" % bits[0])
    filepath = parser.compile_filter(bits[1])
    return SsiNode(filepath, parsed)




from __future__ import absolute_import

from django_compat_patcher.deprecation import *

warnings.warn("django.utils.unittest will be removed in Django 1.9.",
              RemovedInDjango19Warning, stacklevel=2)

try:
    from unittest2 import *
except ImportError:
    from unittest import *




from functools import wraps

from django.utils import six

from django_compat_patcher.deprecation import *

def memoize(func, cache, num_args):
    """
    Wrap a function so that results for any argument tuple are stored in
    'cache'. Note that the args to the function must be usable as dictionary
    keys.

    Only the first num_args are considered when creating the key.
    """
    warnings.warn("memoize wrapper is deprecated and will be removed in "
                  "Django 1.9. Use django.utils.lru_cache instead.",
                  RemovedInDjango19Warning, stacklevel=2)

    @wraps(func)
    def wrapper(*args):
        mem_args = args[:num_args]
        if mem_args in cache:
            return cache[mem_args]
        result = func(*args)
        cache[mem_args] = result
        return result
    return wrapper



#!/usr/bin/env python

import sys, os

os.chdir(os.path.dirname(os.path.abspath(__file__)))  # security

from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read().strip()

classifiers = """\
Development Status :: 4 - Beta
Intended Audience :: Developers
Intended Audience :: Information Technology
Intended Audience :: System Administrators
License :: OSI Approved :: MIT License
Programming Language :: Python
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS :: MacOS X
"""


setup(
    name='django-compat-patcher',
    version=read("VERSION"),
    author='Pascal Chambon & others',
    author_email='pythoniks@gmail.com',
    url='https://github.com/pakal/django-compat-patcher',
    license="http://www.opensource.org/licenses/mit-license.php",
    platforms=["any"],
    description="A monkey-patching system to ease the transition between Django versions.",
    classifiers=filter(None, classifiers.split("\n")),
    long_description=read("README.rst"),

    #package_dir={'': 'src'},
    packages=find_packages(),

    # test_suite='your.module.tests',

    use_2to3=True,
    #convert_2to3_doctests=['src/your/module/README.txt'],
    #use_2to3_fixers=['your.fixers'],
    use_2to3_exclude_fixers=['lib2to3.fixes.fix_import'],
)




from __future__ import absolute_import, print_function, unicode_literals

import functools
import logging
import types
import warnings as stdlib_warnings  # do NOT use elseware than in emit_warning()
from functools import wraps, partial

import sys
from django.utils import six


class SkipFixerException(Exception):
    """
    Exception to signal a fixer which is not 
    applicable in that project context.
    """
    pass


def get_patcher_setting(name, settings=None):
    """
    Fetches the value of the DCP setting.

    If it's not found, a default value is returned for the setting.

    If provided, the 'settings' arguments is a dict which *completely*
    replaces django settings (no fallback occurs).

    :param name: The name of the setting
    :param settings: Possible replacement for the project's settings
    :return: The value of the setting "name"
    """
    from django.conf import settings as django_settings
    from . import default_settings

    if settings is None:
        settings = django_settings

    if not name.startswith("DCP"):
        raise ValueError("Only 'DCP_XXX' setting names are allowed in get_patcher_setting()")

    try:
        if isinstance(settings, dict):
            setting = settings[name]
        else:
            setting = getattr(settings, name)  # Will break if unknown setting
    except (AttributeError, KeyError) as e:
        #print("EXCEPTION IN get_patcher_setting", name, e)
        setting = getattr(default_settings, name)

    # Micromanaging, because a validation Schema is overkill for now
    if name == "DCP_LOGGING_LEVEL":
        assert setting is None or hasattr(logging, setting), repr(setting)
    elif name in ("DCP_PATCH_INJECTED_OBJECTS", "DCP_ENABLE_WARNINGS"):
        assert isinstance(setting, bool), repr(setting)
    else:
        assert (setting == "*" or
                (isinstance(setting, list) and all(isinstance(f, six.string_types) for f in setting))), setting
    return setting


def apply_runtime_settings(settings):
    """
    Change at runtime the logging/warnings settings.
    """
    global DCP_ENABLE_WARNINGS, DCP_LOGGING_LEVEL, _initial_setup_done

    settings = settings or {}

    if not _initial_setup_done:
        DCP_ENABLE_WARNINGS = get_patcher_setting("DCP_ENABLE_WARNINGS")
        DCP_LOGGING_LEVEL = get_patcher_setting("DCP_LOGGING_LEVEL")
        _initial_setup_done = True

    if "DCP_ENABLE_WARNINGS" in settings:
        dcp_enable_warnings = settings["DCP_ENABLE_WARNINGS"]
        assert dcp_enable_warnings in (True, False), dcp_enable_warnings
        DCP_ENABLE_WARNINGS = dcp_enable_warnings  # runtime switch on/off

    if "DCP_LOGGING_LEVEL" in settings:
        dcp_logging_level = settings["DCP_LOGGING_LEVEL"]
        assert dcp_logging_level is None or hasattr(logging, dcp_logging_level), dcp_logging_level
        DCP_LOGGING_LEVEL = dcp_logging_level


# lazy setup of the following settings
_initial_setup_done = False

# global on/off switch for (deprecation) warnings, to be modified by patch() if wanted
DCP_ENABLE_WARNINGS = False

# global logging level string, or None if no logging
DCP_LOGGING_LEVEL = None


def emit_log(message, level="INFO"):
    if DCP_LOGGING_LEVEL is None:
        return
    if getattr(logging, level) < getattr(logging, DCP_LOGGING_LEVEL):
        return
    full_message = "[DCP_%s] %s" % (level, message)
    print(full_message, file=sys.stderr)


def emit_warning(message, category=None, stacklevel=1):
    category = category or DeprecationWarning
    assert issubclass(category, DeprecationWarning), category  # only those are used atm
    if DCP_ENABLE_WARNINGS:
        stdlib_warnings.warn(message, category, stacklevel + 1)


def get_django_version():
    import django
    return django.get_version()


def skip_if_app_not_installed(app_name):
    """
    Raises a SkipFixerException if app_name is not enabled in Django settings.
    """
    from django.conf import settings
    if app_name not in settings.INSTALLED_APPS:
        raise SkipFixerException("%s is not enabled in INSTALLED_APPS" % app_name)


def _patch_injected_object(object_to_patch):
    # we expect a "custom" object here...
    assert object_to_patch not in (True, False, None)
    if get_patcher_setting('DCP_PATCH_INJECTED_OBJECTS'):
        #print("PATCHING injected object", object_to_patch)
        try:
            setattr(object_to_patch, "__dcp_injected__", True)
        except AttributeError:
            pass # '@properties' and such special objects can't be modified


def _is_simple_callable(obj):
    return isinstance(obj, (types.FunctionType, types.BuiltinFunctionType, functools.partial))


def inject_attribute(target_object, target_attrname, attribute):
    """
    :param target_object: The object to patch
    :param target_attrname: The name given to the new attribute in the object to patch
    :param attribute: The attribute to inject: must not be a callable
    """
    assert attribute is not None
    assert not _is_simple_callable(attribute), attribute
    assert not isinstance(attribute, six.class_types), attribute

    _patch_injected_object(attribute)
    setattr(target_object, target_attrname, attribute)


def inject_callable(target_object, target_callable_name, patch_callable):
    """
    :param target_object: The object to patch
    :param target_callable_name: The name given to the new callable in the object to patch
    :param patch_callable: The callable to inject: must be a callable, but not a class
    """
    assert _is_simple_callable(patch_callable), patch_callable

    _patch_injected_object(patch_callable)
    setattr(target_object, target_callable_name, patch_callable)


def inject_module(target_module_name, target_module):
    """
    :param target_module_name: The name of the new module in sys.modules
    :param target_module: The new module
    """
    target_module_name = str(target_module_name)  # Python2 compatibility
    assert isinstance(target_module, types.ModuleType), target_module
    assert sys.modules.get(target_module_name) is None

    _patch_injected_object(target_module)

    sys.modules[target_module_name] = target_module


def inject_class(target_object, target_klassname, klass):
    """
    :param target_object: The object to patch
    :param target_klassname: The name given to the new class in the object to patch
    :param klass: The class to inject : must be a class
    """
    assert isinstance(klass, six.class_types), klass

    _patch_injected_object(klass)
    setattr(target_object, target_klassname, klass)


def inject_callable_alias(source_object, source_attrname,
                          target_object, target_attrname):
    """
    Create and inject an alias for the source callable (not a class),
    by handling logging/warnings

    Returns the created alias.
    """
    source_callable = getattr(source_object, source_attrname)
    assert _is_simple_callable(source_callable), source_callable

    @wraps(source_callable)
    def wrapper(*args, **kwds):
        # we dunno if it's a backwards or forwards compatibility shim...
        emit_warning('%s.%s, which is an alias for %s.%s, was called. One of these is deprecated.' %
                     (target_object, source_attrname, source_object, source_attrname),
                     DeprecationWarning)
        return source_callable(*args, **kwds)

    _patch_injected_object(wrapper)
    setattr(target_object, target_attrname, wrapper)

    return wrapper


def register_import_alias(alias_name, real_name):
    from . import import_proxifier
    import_proxifier.install_import_proxifier()  # idempotent activation
    import_proxifier.register_module_alias(alias_name=alias_name,
                                           real_name=real_name)


def _tuplify_version(version):
    """
    Coerces the version string (if not None), to a version tuple.
    Ex. "1.7.0" becomes (1, 7, 0).
    """
    if version is None:
        return version
    if isinstance(version, six.string_types):
        version = tuple(int(x) for x in version.split("."))
    assert len(version) <= 4, version
    assert (1, 3) <= version, version
    assert all(isinstance(x, six.integer_types) for x in version), version
    return version


def _detuplify_version(input_tuple):
    """
    Coerces the version tuple (if not None), to a version string.
    Ex. (1, 7, 0) becomes "1.7.0".
    """
    if input_tuple is None:
        return ""
    assert isinstance(input_tuple, tuple)
    string = ".".join(str(number) for number in input_tuple)
    return string



from __future__ import absolute_import, print_function, unicode_literals

from functools import partial

from django.core.exceptions import ImproperlyConfigured
from django.utils import six

from ..deprecation import *
from ..registry import register_compatibility_fixer

# for backward-compatibility fixers
django1_10_bc_fixer = partial(register_compatibility_fixer,
                              fixer_reference_version="1.10",
                              fixer_applied_from_django="1.10")


def _get_url_utils():
    try:
        from django.urls import get_callable, RegexURLPattern, RegexURLResolver, NoReverseMatch
    except ImportError:
        from django.core.urlresolvers import get_callable, RegexURLPattern, RegexURLResolver, NoReverseMatch   # old location
    return get_callable, RegexURLPattern, RegexURLResolver, NoReverseMatch



@register_compatibility_fixer(fixer_reference_version="1.10", fixer_applied_upto_django="1.10")
def fix_incoming_django_urls(utils):
    """
    Put a forward compatibility import path for django.urls, which replaces django.core.urlresolvers
    """
    from django.core import urlresolvers
    utils.inject_module("django.urls", urlresolvers)


@django1_10_bc_fixer()
def fix_deletion_templatetags_future(utils):
    """
    Preserve the "future" templatetags library, with its improved `firstof` and `cycle` tags.
    """
    import django.templatetags
    from ..django_legacy.django1_10.templatetags import future
    utils.inject_module("django.templatetags.future", future)
    utils.inject_attribute(django.templatetags, "future", future)

    from django.template.backends import django as django_templates
    _old_get_installed_libraries = django_templates.get_installed_libraries

    def get_installed_libraries():
        libraries = _old_get_installed_libraries()  # tries real __import__() calls on submodules
        libraries["future"] = "django.templatetags.future"
        # print(">>>>> FINAL libraries", libraries)
        return libraries

    utils.inject_callable(django_templates, "get_installed_libraries", get_installed_libraries)


@django1_10_bc_fixer()
def fix_deletion_template_defaulttags_ssi(utils):
    """
    Preserve the "ssi" default template tag.
    """
    import django.template.defaulttags
    from ..django_legacy.django1_10.template import defaulttags

    utils.inject_callable(django.template.defaulttags, "include_is_allowed", defaulttags.include_is_allowed)
    utils.inject_class(django.template.defaulttags, "SsiNode", defaulttags.SsiNode)
    utils.inject_callable(django.template.defaulttags, "ssi", defaulttags.ssi)
    django.template.defaulttags.register.tag(defaulttags.ssi)

    from django.template.engine import Engine
    _old_init = Engine.__init__

    def __init__(self, dirs=None, app_dirs=False, allowed_include_roots=None, **kwargs):
        if allowed_include_roots is None:
            allowed_include_roots = []
        if isinstance(allowed_include_roots, six.string_types):
            raise ImproperlyConfigured(
                "allowed_include_roots must be a tuple, not a string.")
        self.allowed_include_roots = allowed_include_roots
        _old_init(self, dirs=dirs, app_dirs=app_dirs, **kwargs)

    utils.inject_callable(Engine, "__init__", __init__)


@django1_10_bc_fixer()
def fix_behaviour_urls_resolvers_RegexURLPattern(utils):
    """
    Restore support for dotted string view in RegexURLPattern,
    instead of view object.
    """

    get_callable, RegexURLPattern, RegexURLResolver, NoReverseMatch = _get_url_utils()

    @property
    def callback(self):
        callback = self.__dict__["callback"]  # bypass descriptor
        if isinstance(callback, six.string_types):
            callback_obj = get_callable(callback)
        else:
            callback_obj = callback
        return callback_obj

    @callback.setter
    def callback(self, value):
        self.__dict__["callback"] = value  # bypass descriptor

    # we inject a DATA-DESCRIPTOR, so it'll be accessed in prority
    # over "self.callback" instance attribute
    utils.inject_attribute(RegexURLPattern, "callback", callback)

    def add_prefix(self, prefix):
        """
        Adds the prefix string to a string-based callback.
        """
        callback = self.__dict__["callback"]
        if not prefix or not isinstance(callback, six.string_types):
            return
        self.callback = prefix + '.' + callback
    utils.inject_callable(RegexURLPattern, "add_prefix", add_prefix)

    original_lookup_str = RegexURLPattern.lookup_str
    @property  # not cached...
    def lookup_str(self):
        callback = self.__dict__["callback"]
        if isinstance(callback, six.string_types):
            # no need for warning, already emitted above
            return callback  # already a dotted path to view
        return original_lookup_str.__get__(self, self.__class__)
    utils.inject_attribute(RegexURLPattern, "lookup_str", lookup_str)


@django1_10_bc_fixer()
def fix_behaviour_core_urlresolvers_reverse_with_prefix(utils):
    """
    Preserve the ability to call urlresolver on dotted string view,
    instead of explicit view name.
    """
    from django.utils.functional import cached_property
    get_callable, RegexURLPattern, RegexURLResolver, NoReverseMatch = _get_url_utils()

    original_reverse_with_prefix = RegexURLResolver._reverse_with_prefix

    def _reverse_with_prefix(self, lookup_view, _prefix, *args, **kwargs):
        original_lookup = lookup_view
        try:
            if self._is_callback(lookup_view):
                utils.emit_warning(
                    'Reversing by dotted path is deprecated (%s).' % original_lookup,
                    RemovedInDjango110Warning, stacklevel=3
                )
                lookup_view = get_callable(lookup_view)
        except (ImportError, AttributeError) as e:
            raise NoReverseMatch("Error importing '%s': %s." % (lookup_view, e))
        return original_reverse_with_prefix(self, lookup_view, _prefix, *args, **kwargs)
    utils.inject_callable(RegexURLResolver, "_reverse_with_prefix", _reverse_with_prefix)



@django1_10_bc_fixer()
def fix_behaviour_conf_urls_url(utils):
    """
    Support passing views to url() as dotted strings instead of view objects.
    """
    get_callable, RegexURLPattern, RegexURLResolver, NoReverseMatch = _get_url_utils()

    from django.conf import urls

    def url(regex, view, kwargs=None, name=None, prefix=''):
        if isinstance(view, (list, tuple)):
            # For include(...) processing.
            urlconf_module, app_name, namespace = view
            return RegexURLResolver(regex, urlconf_module, kwargs, app_name=app_name, namespace=namespace)
        else:
            if isinstance(view, six.string_types):
                utils.emit_warning(
                    'Support for string view arguments to url() is deprecated and '
                    'will be removed in Django 1.10 (got %s). Pass the callable '
                    'instead.' % view,
                    RemovedInDjango110Warning, stacklevel=2
                )
                if not view:
                    raise ImproperlyConfigured('Empty URL pattern view name not permitted (for pattern %r)' % regex)
                if prefix:
                    view = prefix + '.' + view
            return RegexURLPattern(regex, view, kwargs, name)
    assert callable(urls.url)
    utils.inject_callable(urls, "url", url)


@django1_10_bc_fixer()
def fix_deletion_conf_urls_patterns(utils):
    """
    Preserve the patterns() builder for django urls.
    """
    from django.core.urlresolvers import RegexURLPattern
    from django.conf import urls

    def patterns(prefix, *args):
        utils.emit_warning(
            'django.conf.urls.patterns() is deprecated and will be removed in '
            'Django 1.10. Update your urlpatterns to be a list of '
            'django.conf.urls.url() instances instead.',
            RemovedInDjango110Warning, stacklevel=2
        )
        pattern_list = []
        for t in args:
            if isinstance(t, (list, tuple)):
                t = urls.url(prefix=prefix, *t)
            elif isinstance(t, RegexURLPattern):
                t.add_prefix(prefix)
            pattern_list.append(t)
        return pattern_list
    utils.inject_callable(urls, "patterns", patterns)

    urls.__all__.append(str("patterns"))  # so that star imports work fine


@django1_10_bc_fixer()
def fix_behaviour_template_smartif_OPERATORS_equals(utils):
    """
    Preserve support for a single '=' sign in {% if %} tag.
    """
    from django.template import smartif
    smartif.OPERATORS['='] = smartif.OPERATORS['==']  # operator alias

 

''' REQUIRES PYTHON >= 3.3
@django1_10_bc_fixer()
def fix_deletion_core_context_processors(utils):
    """
    Keep django.core.context_processors middlewares as aliases for
    those of django.template.context_processors.
    """
    from .. import import_proxifier
    import_proxifier.install_module_alias_finder()  # idempotent
    import_proxifier.register_module_alias(module_alias="django.core.context_processors middlewares",
                                           real_module="django.template.context_processors")
'''
